package hauer.calculator;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.StringTokenizer;

/**
 * Die Logik hinter dem Taschenrechner. Verwaltet den Stack und überimmt die Analyse der Eingabe.
 * @author hauer
 *
 */
public class Calculator {

	/**
	 * Der vom Taschenrechner verwaltete Stapel.
	 */
	private Stack<Operand> stack;

	/**
	 * Ob das Programm im interaktiven Modus weiter laufen soll.
	 */
	private boolean running;

	/**
	 * Erstellt einen neuen Taschenrechner mit einem neuen, leeren Stack.
	 */
	public Calculator() {
		this.running = false;
		this.stack = new Stack<Operand>();
	}
	
	/**
	 * Ließt die Standard Eingabe ein und zeigt einen kleinen Prompt ">" an, um den Nutzer zur Eingabe aufzufordern.
	 * Ruft den Parser auf um die Eingabe in Tokens aufzuspalten.
	 */
	public void runInteractive() {
		Scanner scanner = new Scanner(System.in);
		this.running = true;
		while(running) {
			System.out.print("> ");
			String input = scanner.nextLine();
			parse(input);
		}
		end();
		scanner.close();
	}

	/**
	 * Ließt eine Datei Zeile für Zeile aus und schmeißt die eingelesenen Strings in den Parser.
	 * @param file Die einzulesende Datei.
	 */
	public void parseFile(File file) {
		Scanner scanner = null;
		try {
			scanner = new Scanner(file);
		} catch (FileNotFoundException e) {
			System.out.println("[ERROR] File not found! Exiting!");
			System.exit(16);
		}
		while (scanner.hasNextLine()) {
			String input = scanner.nextLine();
			parse(input);
		}
		end();
		scanner.close();
	}
	
	/**
	 * Spaltet einen String an seinen Leerzeichen auf und ruft für jeden Teilstring die chooseAction()-Methode auf.
	 * @param input Der String, der aufgespalten werden soll.
	 */
	private void parse(String input){
		StringTokenizer tokenizer = new StringTokenizer(input, " ");
		while(tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			chooseAction(token);
		}
	}
	
	/**
	 * Wählt eine Aktion (Operation durchführen oder Operand auf den Stack legen) anhand des übergebenen Tokens aus.
	 * @param token
	 */
	private void chooseAction(String token) {
		switch (token) {
		/* Addition, Subtraktion, Multiplikation und Division mit Lambda-Ausdrücken */
		case "+":
			this.performBinaryOperation((a, b)->new Operand(b.value + a.value));
			break;
		case "-":
			this.performBinaryOperation((a, b) -> new Operand(b.value - a.value));
			break;
		case "*":
			this.performBinaryOperation((a, b) -> new Operand(b.value * a.value));
			break;
		case "/":
			this.performBinaryOperation((a, b) -> new Operand(b.value / a.value));
			break;
		/* Unäre Operatoren */
		case "sqrt":
			this.performUnaryOperation((a)->new Operand(Math.sqrt(a.value)));
		/* Verwaltungsbefehle für den interaktiven Modus */
		case "peek":
			if (this.stack.isEmpty()) {
				System.out.println("[INFO] The stack is empty.");
			} else {
				System.out.println(this.stack.peek());
			}
			break;
		case "pop":
			if (this.stack.isEmpty()) {
				System.out.println("[ERROR] pop: The stack is empty.");
			} else {
				System.out.println(this.stack.pop());
			}
			break;
		case "exit":
			this.running = false;
			break;
		default:
		/* Versuche im Standardfall eine Zahl zu parsen */
			try {
				double newDouble = Double.parseDouble(token);
				Operand newValue = new Operand(newDouble);
				this.stack.push(newValue);
			} catch (NumberFormatException e) {
				System.out.println("[ERROR] Number formatted incorrectly!");
			}
		}
	}
	
	/**
	 *  Führe eine unäre Operation aus: Entferne ein Element vom Stack und lege das Ergebnis zurück.
	 *  
	 *  @param op Der Operator
	 *
	 */
	private void performUnaryOperation(UnaryOperation<Operand> op) {
		if(this.stack.getSize() < 1) {
			System.out.println("[ERROR] Not enough values on the stack!");
			return;
		}
		Operand a = stack.pop();
		Operand b = op.apply(a);
		this.stack.push(b);
	}

	/**
	 * Führe eine binäre Operation aus: Entferne zwei Elemente vom Stack und lege das Ergebnis zurück.
	 *
	 * @param op Der auszuführende Operator
	 */
	private void performBinaryOperation(BinaryOperation<Operand> op) {
		if (this.stack.getSize() < 2) {
			System.out.println("[ERROR] Not enough values on the stack!");
			return;
		}
		Operand a = stack.pop();
		Operand b = stack.pop();
		Operand c = op.apply(a, b);
		this.stack.push(c);
	}
	
	/**
	 * Beendet das Programm. Falls noch etwas auf dem Stack liegt, wird es als finales Ergebnis ausgegeben.
	 */
	private void end() {
		if (!this.stack.isEmpty()) {
			chooseAction("peek");
		}
		System.out.println("Good Bye!");
	}
}
