package hauer.calculator;

/**
 * Die Stapelklasse.
 * 
 * Verwaltet nur das oberste Element des Stapels, der Rest des Stapels ist mit diesem verkettet.
 * Auch bekannt als "Least In, First Out" (LIFO)-Warteschlange.
 * @author hauer
 *
 * @param <T> Was in den Elementen des Stapels gespeichert werden soll.
 */
public class Stack<T> {

	/**
	 * Das oberste Element des Stapels.
	 */
	private Element first;
	
	/**
	 * Die Größe des Stapels.
	 */
	private int size;
	
	public Stack(){
		this.size = 0;
		this.first = null;
	}
	
	/**
	 * Legt ein neues Element auf den Stapel.
	 * @param value
	 */
	public void push(T value) {
		Element newElement = new Element(value, first);
		this.first = newElement;
		this.size++;
	}
	
	/**
	 * Sieht sich das oberste Element auf dem Stapel an.
	 * @return Das oberste Element, oder null wenn es nicht existiert.
	 */
	public T peek() {
		if(this.first != null) {
			return this.first.getValue();
		}
		return null;
	}
	
	/**
	 * Entfernt das oberste Element vom Stapel.
	 * @return Das oberste Element, oder null wenn es nicht existiert.
	 */
	public T pop(){
		if(this.first != null) {
			T result = this.first.getValue();
			this.first = this.first.getNext();
			size--;
			return result;
		}
		return null;
	}
	
	/**
	 * Überprüft, ob der Stapel leer ist.
	 * @return Ob der Stapel leer ist.
	 */
	public boolean isEmpty(){
		return this.first == null;
	}
	
	/**
	 * Gibt die Größe des Stapels zurück.
	 * @return Die Anzahl an Elementen im Stapel.
	 */
	public int getSize() {
		return this.size;
	}
	
	/**
	 * Innere Element-Klasse zum Verwalten der Objekte auf dem Stapel.
	 * @author hauer
	 *
	 */
	private class Element {
		
		/**
		 * Das nächste Element.
		 */
		private Element next;
		
		/**
		 * Der im Element gespeicherte Wert.
		 */
		private final T value;
		
		/**
		 * Erstellt ein neues Element mit einem Wert, ohne Nachfolger.
		 * @param value Der zu speichernde Wert.
		 */
		public Element(T value) {
			this.next = null;
			this.value = value;
		}
		
		/**
		 * Erstellt ein neues Element mit einem Wert und einem Nachfolger.
		 * @param value Der zu speichernde Wert.
		 * @param next Der Nachfolger des Elementes.
		 */
		public Element(T value, Element next) {
			this.value = value;
			this.next = next;
		}
		
		/**
		 * Gibt den gespeicherten Wert zurück.
		 * @return Der Wert.
		 */
		public T getValue() {
			return this.value;
		}
		
		/**
		 * Gibt den Nachfolger des Elementes zurück.
		 * @return Der Nachfolger, oder null wenn es keinen gibt.
		 */
		public Element getNext() {
			return this.next;
		}
		
		/**
		 * Setzt den Nachfolger des Elementes auf das übergebene Element.
		 * @param next Der zukünftige Nachfolger.
		 */
		public void setNext(Element next) {
			this.next = next;
		}
	}
}
